"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.getChangedWorkspaces = getChangedWorkspaces;
exports.matchWorkspaces = matchWorkspaces;

var _path = _interopRequireDefault(require("path"));

var _pkgUp = _interopRequireDefault(require("pkg-up"));

var _gitActions = require("./gitActions");

var _getModularRoot = _interopRequireDefault(require("./getModularRoot"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Gets a list of changed files, then maps them to their workspace and returns a subset of WorkspaceContent
async function getChangedWorkspaces(workspaceContent, targetBranch) {
  const diffedFiles = (0, _gitActions.getDiffedFiles)(targetBranch);
  const modularRoot = (0, _getModularRoot.default)(); // Resolve each of the changed files to their nearest package.json. The resulting list can contain duplicates and null holes

  const packageManifestPaths = await Promise.all(diffedFiles.map(changedFile => (0, _pkgUp.default)({
    cwd: _path.default.dirname(changedFile)
  })));
  return matchWorkspaces(packageManifestPaths, modularRoot, workspaceContent);
} // Match workspace manifest paths to a subset of WorkspaceContent. This function works completely in memory and is test-friendly


function matchWorkspaces(packagePaths, root, workspaceContent) {
  const [packages, workspaces] = workspaceContent;
  const workspaceEntries = Object.entries(workspaces);
  const result = [new Map([]), {}];

  for (const packagePath of packagePaths) {
    const [resultPackages, resultWorkspaces] = result; // Ignore holes

    if (!packagePath) continue; // Get the package directory from the package.json path and make it relative to the root, for comparison with the original WorkspaceContent

    const packageDir = _path.default.relative(root, _path.default.dirname(packagePath)); // Match the package directory to its entry WorkspaceContent, using pathEquality


    const foundEntry = workspaceEntries.find(([_, {
      location
    }]) => pathEquality(location, packageDir)); // If found, insert the entries into the WorkspaceContent that we are building

    if (foundEntry) {
      const [foundWorkspaceName, foundWorkspace] = foundEntry;
      resultWorkspaces[foundWorkspaceName] = foundWorkspace;
      const foundPackage = packages.get(foundWorkspaceName);

      if (foundPackage) {
        resultPackages.set(foundWorkspaceName, foundPackage);
      }
    }
  }

  return result;
} // Path equality !== string equality


function pathEquality(path1, path2) {
  if (path1 === null || path2 === null) return false;
  path1 = _path.default.resolve(path1);
  path2 = _path.default.resolve(path2); // Win32 is case insensitive

  return process.platform === 'win32' ? path1.toLowerCase() === path2.toLowerCase() : path1 === path2;
}
//# sourceMappingURL=getChangedWorkspaces.js.map